import tensorflow as tf

@tf.custom_gradient
def grad_reverse(x, weight):
    y = tf.identity(x)
    def custom_grad(dy):
        return -weight * dy, -weight * dy #lower is better
    return y, custom_grad

class GradReverse(tf.keras.layers.Layer):
    def __init__(self):
        super().__init__()

    def call(self, x, weight):
        return grad_reverse(x, weight)
    
class Behaviour_Decoder(tf.keras.Model):
  def __init__(self, gru_units, factor_num, dropout_rate):
    super(Behaviour_Decoder, self).__init__()

    self.gru = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_units, 
                                    return_sequences=True))
    self.batch_norm = tf.keras.layers.BatchNormalization()
    self.gru2 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_units, 
                                    return_sequences=True))
    self.batch_norm2 = tf.keras.layers.BatchNormalization()
    

    self.inter = tf.keras.layers.Dense(512, activation='relu', 
                                       kernel_regularizer=tf.keras.regularizers.L2(1.0),
                                       )
    self.batch_norm_inter = tf.keras.layers.BatchNormalization()
    self.drop_layer = tf.keras.layers.Dropout(dropout_rate)
    self.readout = tf.keras.layers.Dense(factor_num, use_bias=False)
    
  def call(self, neural_batch, get_cell_activations = False): 
      neural_batch =tf.cast(neural_batch, tf.float32)
      neural_act = self.batch_norm(self.gru(tf.cast(neural_batch, tf.float32)))
      neural_act = self.batch_norm2(self.gru2(neural_act)) 

      self.behaviour_pred = self.drop_layer(self.batch_norm_inter(self.inter(neural_act)))
      self.final = self.readout(self.behaviour_pred)

      return self.final
        
  
class Behaviour_Factor_Decoder(tf.keras.Model):
    def __init__(self, units, output_size, dropout_rate):
        super(Behaviour_Factor_Decoder, self).__init__()
        
        self.interd = tf.keras.layers.Dense(units, activation='relu', 
                                           kernel_regularizer=tf.keras.regularizers.L2(1.0),
                                           )

        self.drop_layerd = tf.keras.layers.Dropout(dropout_rate) 
        self.readoutd = tf.keras.layers.Dense(output_size, use_bias=False)
        
    def call(self, neural_act, get_cell_activations = False): 
        self.behaviour_pred = self.drop_layerd(self.interd(neural_act))
        self.final = self.readoutd(self.behaviour_pred)
        return self.final
    
class Neural_Model_noBN_Enc(tf.keras.Model):
  def __init__(self, gru_units, output_size, dropout_rate, final_activation):
    super(Neural_Model_noBN_Enc, self).__init__()
    
    self.gru = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_units, 
                                    kernel_regularizer=tf.keras.regularizers.L2(1000.0), 
                                    recurrent_regularizer = tf.keras.regularizers.L2(1000.0),
                                    recurrent_dropout= 0.2,
                                    return_sequences=True))

    self.gru2 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_units, 
                                    kernel_regularizer=tf.keras.regularizers.L2(1000.0), 
                                    recurrent_regularizer = tf.keras.regularizers.L2(1000.0),
                                    recurrent_dropout= 0.2,
                                    return_sequences=True))

    self.gru3 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_units,
                                    kernel_regularizer=tf.keras.regularizers.L2(1000.0),
                                    recurrent_regularizer = tf.keras.regularizers.L2(1000.0),
                                    recurrent_dropout= 0.2,
                                    return_sequences=False))


    self.inter = tf.keras.layers.Dense(512, activation='relu',kernel_regularizer=tf.keras.regularizers.L2(1000))
    
    self.drop_layer = tf.keras.layers.Dropout(dropout_rate)
    self.readout = tf.keras.layers.Dense(output_size, use_bias=False, activation=final_activation)

    
  def call(self, neural_batch, get_cell_activations = False): 
      neural_batch =tf.convert_to_tensor(tf.cast(neural_batch, tf.float32))
      neural_act = self.gru(neural_batch)
      neural_act = self.gru2(neural_act)
      neural_act = self.gru3(neural_act)

      self.behaviour_pred = self.drop_layer(self.inter(neural_act))
      self.final = self.readout(self.behaviour_pred)

      return self.final
    
class Neural_Model_noBN_Dec(tf.keras.Model):
  def __init__(self, gru_units):
    super(Neural_Model_noBN_Dec, self).__init__()
    
    self.gru = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_units, 
                                    kernel_regularizer=tf.keras.regularizers.L2(0.1), 
                                    recurrent_regularizer = tf.keras.regularizers.L2(0.1),
                                    return_sequences=True))

  def call(self, neural_batch, get_cell_activations = False): 
      neural_batch =tf.cast(neural_batch, tf.float32)
      neural_act = self.gru(neural_batch)

      return neural_act
  
    
class Factor_Decoder(tf.keras.Model):
    def __init__(self, units, output_size, dropout_rate):
        super(Factor_Decoder, self).__init__()
        
        self.inter = tf.keras.layers.Dense(units, activation='relu', 
                                           kernel_regularizer=tf.keras.regularizers.L2(10),
                                           )

        self.drop_layer = tf.keras.layers.Dropout(dropout_rate) 
        self.readout = tf.keras.layers.Dense(output_size, use_bias=False)
        
    def call(self, neural_act, get_cell_activations = False): 
        self.behaviour_pred = self.drop_layer(self.inter(neural_act))
        self.final = self.readout(self.behaviour_pred)
        return self.final
    
    

class SABLE(tf.keras.Model):
    def __init__(self, latent_dim, output_size, num_factors, num_steps):
        super(SABLE, self).__init__()
        self.num_steps = num_steps
        self.latent_dim = latent_dim
        self.grad_rev = GradReverse()
        self.behaviour_decoder = Behaviour_Decoder(256, num_factors,0.05)
        self.behaviour_factor_decoder = Behaviour_Factor_Decoder(num_factors, 2, 0.1)
        self.encoder = Neural_Model_noBN_Enc(512, latent_dim*2, 0.8, None)
        self.decoder = Neural_Model_noBN_Dec(256)
        self.rates = Factor_Decoder(128, output_size, 0.2)
        

    def encode(self, x, weight = None, reverse_gradient=False, training=False):
      if reverse_gradient:
          out = self.grad_rev(self.encoder(x, training=training), weight)
      else:
          out = self.encoder(x, training=training)
      mean, logvar = tf.split(out, num_or_size_splits=2, axis=1)
      return mean, logvar
  
    def reparameterize(self, mean, logvar):
      eps = tf.random.normal(shape=mean.shape)
      return eps * tf.exp(logvar * .5) + mean
  
    def decode(self, z_all, training=False):      
      
      logits = self.decoder(z_all, training=training)   
      return logits

