import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
import h5py
import time
from sklearn.manifold import TSNE
from models import *
from functions import *
import pathlib
from tensorflow.python.framework import ops

physical_devices = tf.config.list_physical_devices('GPU') 
tf.config.experimental.set_memory_growth(physical_devices[0], True)
plt.rcParams["figure.figsize"] = (5,5)

train_data, test_data, train_target, test_target, train_behaviours, test_behaviours = load_data(train_files, test_files)

num_epochs = 500
batch_size = 64
optimizer_b = tf.keras.optimizers.Adam(0.0001)
optimizer_s = tf.keras.optimizers.Adam(0.0000001)
optimizer_kl = tf.keras.optimizers.Adam(0.000002)
num_factors = 64
rnn_units = 256
output_shape = 2 #8 for direction, 2 for behaviour pos
latent_dim= 64
output_size = train_data.shape[2]
model = SABLE(latent_dim, output_size, num_factors, train_data.shape[1])


train_losses, valid_losses, train_means, valid_means = [], [], [], []
iterations = 0        
batch_indices = list(range(train_data.shape[0]))
acc = float('inf')
epoch = 0
colors = plt.cm.nipy_spectral(np.arange(8)/8)

all_directions_dict = train_target
direction_index_train = [sorted(set(all_directions_dict)).index(i) for i in train_target]
direction_index_test = [sorted(set(all_directions_dict)).index(i) for i in test_target]
while epoch < num_epochs:
  train_behaviour_decoder = True #if acc < 30000 else False
  kl_weight = kl_weighting(0.01, 1000000.0, 0.2, epoch)
  rev_weight = rev_weightinig(1.0, 0.00000001, 0.01 , epoch)

  for i in range(train_data.shape[0]//batch_size):
      batch = np.random.choice(batch_indices, batch_size, replace = False)
      
      train_step(model, train_data[batch], train_behaviours[batch], kl_weight, rev_weight, optimizer_b, optimizer_s, optimizer_kl, train_behaviour_decoder)
  loss_train = tf.keras.metrics.Mean()
  tl, acc, b_train, kl_train, r2_train = compute_loss(model, train_data, train_behaviours, train_behaviour_decoder, latent_dim, batch_size, rnn_units, num_factors)
  loss_train(tl)
  
  elbo_train = loss_train.result()
  
  loss_test = tf.keras.metrics.Mean()
  tlt, acct, b_test, kl_test, r2_test = compute_loss(model, test_data, test_behaviours, train_behaviour_decoder, latent_dim, test_data.shape[0], rnn_units, num_factors)
  loss_test(tlt)
  elbo_test = loss_test.result()
  print('withREV epoch: {}, ELBO: {:.1f}, {:.1f}, acc: {:.1f}, {:.1f}, behaviour: {:.1f}, {:.1f}, kl: {:.1f}, {:.1f}, r2: {:.3f}, {:.3f}, ws: (kl, rev) {}, {}'
        .format(epoch, elbo_train, elbo_test, acc, acct, b_train, b_test, kl_train, kl_test, r2_train, r2_test, kl_weight, rev_weight))
  with open('output.txt', "a") as text_file:
      text_file.write('epoch: {}, ELBO: {:.1f}, {:.1f}, acc: {:.1f}, {:.1f}, behaviour: {:.1f}, {:.1f}, kl: {:.1f}, {:.1f}, r2: {:.3f}, {:.3f}, ws: (kl, rev) {}, {} \n'
        .format(epoch, elbo_train, elbo_test, acc, acct, b_train, b_test, kl_train, kl_test, r2_train, r2_test, kl_weight, rev_weight))
  if epoch % 20 == 0:
         plot_directions(model, train_data, train_behaviours, train_target, 'train',latent_dim, batch_size, num_factors)
         plot_directions(model, test_data, test_behaviours, test_target, 'test', latent_dim, batch_size, num_factors)
         z_train = np.zeros((train_data.shape[0], latent_dim))
  
         for chunk in range(train_data.shape[0]//batch_size):                
             mean, logvar = model.encode(train_data[chunk*batch_size: (chunk + 1)*(batch_size)])
             z_train[chunk*batch_size: (chunk + 1)*(batch_size)] = model.reparameterize(mean, logvar)
         
         z_test = np.zeros((test_data.shape[0], latent_dim))

         for chunk in range(test_data.shape[0]//(batch_size//2)):                
             mean, logvar = model.encode(test_data[chunk*(batch_size//2): (chunk + 1)*(batch_size//2)])
             z_test[chunk*(batch_size//2): (chunk + 1)*(batch_size//2)] = model.reparameterize(mean, logvar)

         z = np.concatenate((z_train, z_test))
         z_embedded = TSNE(n_components=2).fit_transform(z)
         for t in range(train_behaviours.shape[0]):
             plt.scatter(z_embedded[t,0], z_embedded[t,1], s=2, color=colors[direction_index_train[t]], alpha=0.7)
         for t in range(test_data.shape[0]):
             plt.scatter(z_embedded[train_data.shape[0] + t,0], z_embedded[train_data.shape[0] + t,1], s=7, color=colors[direction_index_test[t]], marker='^')
         plt.savefig('tsne.pdf')
         plt.clf()
  model.save_weights('saved_models/dalfads')
  epoch += 1
